#include <STC8H.H>
#include <string.h>
#include <ctype.h>

#define SYS_FREQ 11059200UL  // System frequency

// Hardware pins
sbit PIN_MRS = P2^7;    // Morse output
sbit PIN_LED = P2^6;    // Status LED
sbit PIN_TXD = P3^1;    // UART TX

// Timing parameters (20wpm)
#define TIME_BASE 60     // Base unit 60ms
#define DOT_TIME  (1 * TIME_BASE)
#define DASH_TIME (3 * TIME_BASE)
#define GAP_SYMBOL (1 * TIME_BASE)
#define GAP_CHAR   (3 * TIME_BASE - GAP_SYMBOL)
#define GAP_WORD   (7 * TIME_BASE)
#define DELAY_LOOP 2000

// Morse code structure
struct MorseData_t {
    char chr;
    const char code *pat;
};

// Morse code table (CODE segment)
code const struct MorseData_t m_table1[] = {
    {'A',".-"},{'B',"-..."},{'C',"-.-."},{'D',"-.."},
    {'E',"."},{'F',"..-."},{'G',"--."},{'H',"...."},
    {'I',".."},{'J',".---"},{'K',"-.-"},{'L',".-.."},
    {'M',"--"},{'N',"-."},{'O',"---"},{'P',".--."},
    {'Q',"--.-"},{'R',".-."},{'S',"..."},{'T',"-"},
    {'U',"..-"}, {'V',"...-"},{'W',".--"},{'X',"-..-"},
    {'Y',"-.--"},{'Z',"--.."},
    {'0',"-----"},{'1',".----"},{'2',"..---"},{'3',"...--"},
    {'4',"....-"},{'5',"....."},{'6',"-...."},{'7',"--..."},
    {'8',"---.."},{'9',"----."},{' ',""}
};

// Function prototypes
void delay_1ms(unsigned int t1);
void timer0_init1(void);
void uart_init1(void);
void uart_send1(char c1);
void mrs_send_sym(char s1);
void mrs_send_chr(char c2);
const char code* get_code1(char c3);

// Precision delay (11.0592MHz optimized)
void delay_1ms(unsigned int t1) {
    unsigned int i1, j1;
    for(i1 = t1; i1 > 0; i1--)
        for(j1 = 108; j1 > 0; j1--);
}

// Timer0 init (500Hz tone)
void timer0_init1() {
    TMOD &= 0xF0;
    TMOD |= 0x02;       // Mode 2
    TH0 = 0xCE;         // 200us interrupt
    TL0 = 0xCE;
    ET0 = 1;            // Enable interrupt
    TR0 = 1;
    EA = 1;
}

// Timer0 ISR (LED sync)
void timer0_isr1() interrupt 1 {
    static unsigned char cnt1 = 0;
    if(++cnt1 >= 5) {    // 1ms period
        PIN_MRS = ~PIN_MRS;
        PIN_LED = ~PIN_LED;
        cnt1 = 0;
    }
}

// UART init (9600bps)
void uart_init1() {
    P_SW1 &= 0x3F;      // Select P3.0/P3.1
    TMOD &= 0x0F;
    TMOD |= 0x20;       // Timer1 mode2
    TH1 = 0xFD;         // 9600 baud
    TL1 = 0xFD;
    SCON = 0x50;
    TR1 = 1;
}

// UART send char
void uart_send1(char c1) {
    SBUF = c1;
    while(!TI);
    TI = 0;
}

// Get morse pattern
const char code* get_code1(char c3) {
    unsigned char idx1;
    c3 = toupper(c3);
    for(idx1=0; idx1<sizeof(m_table1)/sizeof(m_table1[0]); idx1++) {
        if(m_table1[idx1].chr == c3) {
            return m_table1[idx1].pat;
        }
    }
    return NULL;
}

// Send symbol (dot/dash)
void mrs_send_sym(char s1) {
    unsigned int t2 = (s1 == '.') ? DOT_TIME : DASH_TIME;
    
    ET0 = 1;            // Enable tone
    delay_1ms(t2);
    ET0 = 0;            // Disable tone
    PIN_MRS = 0;        // Force low
    PIN_LED = 0;
    delay_1ms(GAP_SYMBOL);
}

// Send full character
void mrs_send_chr(char c2) {
    const char code *ptr1 = get_code1(c2);
    if(!ptr1) return;
    
    while(*ptr1) {
        mrs_send_sym(*ptr1);
        ptr1++;
    }
    delay_1ms(GAP_CHAR);
}

// Main program
void main() {
    const char msg1[] = "CQ CQ CQ THIS IS BG2ABJ CQ AND STAND BY";
    unsigned char n1;
    bit flag1 = 0;
    
    // P2.6/P2.7 push-pull
    P2M1 &= ~0xC0;
    P2M0 |= 0xC0;
    
    timer0_init1();
    uart_init1();
    PIN_MRS = 0;
    PIN_LED = 0;

    while(1) {
        for(n1 = 0; msg1[n1]; n1++) {
            uart_send1(msg1[n1]);  // Send original char
            
            if(msg1[n1] == ' ') {
                if(flag1) {
                    delay_1ms(GAP_WORD);
                    flag1 = 0;
                }
            } else {
                if(!flag1) flag1 = 1;
                mrs_send_chr(msg1[n1]);
            }
        }
        uart_send1('\r');
        uart_send1('\n');
        delay_1ms(DELAY_LOOP);
        flag1 = 0;
    }
}